clear all; close all;

% align images, detect obstruction, mask obstructed areas
% references: https://sites.google.com/site/obstructionfreephotography/
%             https://github.com/davlia-projects/Obstruction-Free-Photography

% load images
count = 5;
im_names = cell(count, 1);
for k = 1:count
    im_names{k} = strcat('data/hanoi_input_', num2str(k), '.png');
end

% construct light field focal stack by reading in (grayscale) images
lightField = zeros([size(imread(im_names{1})), 6]);
for k = 1:count 
    im = im2double(imread(im_names{k}));
    lightField(:,:,:,k) = im;    
end
[m, n, c] = size(squeeze(lightField(:,:,:,1)));

% ues the center layer of the focal stack as reference frame
ref_frame = round(count/2);

% edge detection on reference frame
im1 = im2double(rgb2gray(lightField(:,:,:,ref_frame)));
im1_edge = double(edge(im1,'Canny', 0.1));

% get translation for every frame (excluding reference)
for k = 1:count
    if k ~= ref_frame
        % edge detection on current frame
        im2 = rgb2gray(lightField(:,:,:,k));
        im2_edge=double(edge(im2,'Canny', 0.1));

        % Lucas-Kanade method for optical flow
        % referenced from: https://www.mathworks.com/matlabcentral/fileexchange/48744-lucas-kanade-tutorial-example-1/content/LucasKanadeExample1/html/LKExample1.html
        
        % window size
        ww = 40;
        w = round(ww/2);
        
        % scale image down
        scale_factor = 2;
        corners = corner(imresize(im2_edge, 1/scale_factor));
        corners = corners*scale_factor;
        
        % Discard coners near the margin of the image
        it = 1;
        for i = 1:size(corners,1)
            x_i = corners(i, 2);
            y_i = corners(i, 1);
            if x_i-w>=1 && y_i-w>=1 && ... 
                    x_i+w<=size(im1_edge,1)-1 && ...
                    y_i+w<=size(im1_edge,2)-1
                C(it,:) = corners(i,:);
                it = it+1;
            end
        end

        % visualize edges and corners
        figure();
        imshow(im2_edge);
        hold on
        plot(C(:,1), C(:,2), 'r*');

        % applying the Lucas-Kanade method
        Ix_m = conv2(im1_edge,[-1 1; -1 1], 'valid'); % partial on x
        Iy_m = conv2(im1_edge, [-1 -1; 1 1], 'valid'); % partial on y
        It_m = conv2(im1_edge, ones(2), 'valid') + conv2(im2_edge, -ones(2), 'valid'); % partial on t
        u = zeros(length(C),1);
        v = zeros(length(C),1);
        
        % within window ww * ww
        for it = 1:length(C(:,2))
            i = C(it,2);
            j = C(it,1);
            Ix = Ix_m(i-w:i+w, j-w:j+w);
            Iy = Iy_m(i-w:i+w, j-w:j+w);
            It = It_m(i-w:i+w, j-w:j+w);
            
            % calculate A, b, and nu
            Ix = Ix(:);
            Iy = Iy(:);
            b = -It(:); 
            
            A = [Ix Iy];
            nu = pinv(A)*b;
            
            u(it) = nu(1);
            v(it) = nu(2);
        end;
        
        % visualize optical flow
        figure();
        imshow(im2_edge);
        hold on;
        quiver(C(:,1), C(:,2), u,v, 1,'r')
        

        % get the motion vector for translating the image
        transition = [u,v];
        [row, col] = size(transition);
        
        % min & max euclidean distance between each correspondence
        dist = [];
        for i = 1:row
            dist = [ dist ; sqrt((transition(i,1))^2+(transition(i,2))^2)];
        end
        [distance_min,row] = min(dist);
        [distance_max,row_max] = max(dist);
        
        Vmin = [transition(row, 1) transition(row, 2)];
        Vobstr = [transition(row_max, 1) transition(row_max, 2)];
        im2_translate = imtranslate(im2, -Vmin);
        im2_translate_obstr = imtranslate(im2, -Vobstr);

        % finally, extract the mask for the obstruction
        background = ones(m,n);
        obstruction = zeros(m,n);
        for a = 1:m
            for b = 1:n
                background(a,b) = min(background(a,b), im2_translate(a,b));
                obstruction(a,b) = im1(a,b) - background(a,b);
                
                % if difference in intensity of input and extracted
                % background is below threshold, set as background
                if obstruction(a,b) < 0.2 
                    obstruction(a,b) = 0;
                    else obstruction(a,b) = 1;
                end
                
            end
        end
        
        figure; imshowpair(uint8(background), obstruction, 'montage');
        
        masked = im2;
        masked(obstruction == 1) = NaN;
        imwrite(masked, ['output/hanoi_mask_' num2str(k) '.png']);
    end
end

% calculate mask for reference frame using other images
average = mean(rgb2gray(lightField(:,:,:,ref_frame)), 3);
average_mask = rgb2gray(lightField(:,:,:,ref_frame));
average_mask(abs(average_mask - average) > 0.5) = NaN;

imwrite(masked, ['output/hanoi_mask_' num2str(ref_frame) '.png']);
